! Dean Townsley 2012
!
! init block for helium strip detonation

subroutine Simulation_initBlock(blockID)
  
  use Simulation_data
  use Grid_interface, ONLY : Grid_getBlkIndexLimits, &
    Grid_putPointData, Grid_getCellCoords
  use Eos_interface, ONLY : Eos

  implicit none
#include "Flash.h"
#include "constants.h"
#include "Eos.h"

  
  integer, intent(in) :: blockID

  integer :: i, j, k

  integer, dimension(2,MDIM) :: blkLimits, blkLimitsGC
  integer, dimension(MDIM) :: cell

  real :: enuc

  real, dimension(EOS_NUM) :: eosData
  real, dimension(NSPECIES) :: abundData

  real, allocatable, dimension(:) :: iCoords, jCoords, kCoords
  integer :: isizeGC, jsizeGC, ksizeGC
  real :: ign_dist


!==============================================================================

  ! get essential info about this block
  call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)

  isizeGC = blkLimitsGC(HIGH,IAXIS)
  allocate(iCoords(isizeGC))
  jsizeGC = blkLimitsGC(HIGH,JAXIS)
  allocate(jCoords(jsizeGC))
  ksizeGC = blkLimitsGC(HIGH,KAXIS)
  allocate(kCoords(ksizeGC))
  call Grid_getCellCoords(IAXIS,blockID,CENTER,.true.,iCoords,isizeGC)
  call Grid_getCellCoords(JAXIS,blockID,CENTER,.true.,jCoords,jsizeGC)
  call Grid_getCellCoords(KAXIS,blockID,CENTER,.true.,kCoords,ksizeGC)


  ! loop over all zones and init
  do k = blkLimits(LOW,KAXIS), blkLimits(HIGH,KAXIS)
     do j = blkLimits(LOW,JAXIS), blkLimits(HIGH,JAXIS)
        do i = blkLimits(LOW,IAXIS), blkLimits(HIGH,IAXIS)

           ! select local state based on whether cell is in the strip of fuel
           ! strip runs along the x=0 line (reflection symmetry about x=0)
           if (iCoords(i) <= sim_he_half_width) then
              eosData(:) = sim_fuel%eosData(:)
              abundData(:) = sim_fuel%abundData(:)
           else
              eosData(:) = sim_inert%eosData(:)
              abundData(:) = sim_inert%abundData(:)
           endif

           ! now check for being in ignition, if so will need to reset temperature & thermal state
           ign_dist = (iCoords(i)-sim_ign_x)**2
           if (NDIM>=2) ign_dist = ign_dist + (jCoords(j)-sim_ign_y)**2
           if (NDIM==3) ign_dist = ign_dist + (kCoords(k)-sim_ign_z)**2
           ign_dist = sqrt(ign_dist)

           ! enuc is "pre-set" to trigger energy-generation refinement in ignition region
           enuc = 0.0
           if (ign_dist <= sim_ign_r ) then
             eosData(EOS_TEMP) = max(eosData(EOS_TEMP), sim_ign_temp_center - &
                            (sim_ign_temp_center-sim_ign_temp_edge)*ign_dist/sim_ign_r )
             enuc = 1.1*sim_refNogenEnucThresh
             call Eos(MODE_DENS_TEMP, 1, eosData, abundData)
           endif

           cell(IAXIS) = i
           cell(JAXIS) = j
           cell(KAXIS) = k
           call Grid_putPointData(blockId, CENTER, DENS_VAR, EXTERIOR, cell, eosData(EOS_DENS))
           call Grid_putPointData(blockId, CENTER, TEMP_VAR, EXTERIOR, cell, eosData(EOS_TEMP))
           call Grid_putPointData(blockId, CENTER, HE4_SPEC, EXTERIOR, cell, abundData(HE4_SPEC-SPECIES_BEGIN+1))
           call Grid_putPointData(blockId, CENTER, C12_SPEC, EXTERIOR, cell, abundData(C12_SPEC-SPECIES_BEGIN+1))
           call Grid_putPointData(blockId, CENTER, O16_SPEC, EXTERIOR, cell, abundData(O16_SPEC-SPECIES_BEGIN+1))
           call Grid_putPointData(blockId, CENTER, NI56_SPEC, EXTERIOR, cell, abundData(NI56_SPEC-SPECIES_BEGIN+1))

           call Grid_putPointData(blockId, CENTER, ENUC_VAR, EXTERIOR, cell, enuc)

           call Grid_putPointData(blockId, CENTER, VELX_VAR, EXTERIOR, cell, 0.0)
           call Grid_putPointData(blockId, CENTER, VELY_VAR, EXTERIOR, cell, 0.0)
           call Grid_putPointData(blockId, CENTER, VELZ_VAR, EXTERIOR, cell, 0.0)

           call Grid_putPointData(blockId, CENTER, ENER_VAR, EXTERIOR, cell, eosData(EOS_EINT))
           call Grid_putPointData(blockId, CENTER, EINT_VAR, EXTERIOR, cell, eosData(EOS_EINT))
           call Grid_putPointData(blockId, CENTER, PRES_VAR, EXTERIOR, cell, eosData(EOS_PRES))
           call Grid_putPointData(blockId, CENTER, GAMC_VAR, EXTERIOR, cell, eosData(EOS_GAMC))
           call Grid_putPointData(blockId, CENTER, GAME_VAR, EXTERIOR, cell, &
                                    eosData(EOS_PRES)/(eosData(EOS_DENS)*eosData(EOS_EINT))+1.0)
        enddo

     enddo
  enddo
  
  deallocate(iCoords)
  deallocate(jCoords)
  deallocate(kCoords)

  return
  
end subroutine Simulation_initBlock
